//
// See README-LCALS_license.txt for access and distribution restrictions
//

//
// Header file defining macros, routines, structures used in Loop Subset A.
//

#ifndef SubsetDataA_HXX
#define SubsetDataA_HXX

#define NDPTRSET(v,v0,v1,v2,v3,v4,v5,v6,v7)  \
   v0 = v ;   \
   v1 = v0 + 1 ;  \
   v2 = v0 + domain.jp ; \
   v3 = v1 + domain.jp ; \
   v4 = v0 + domain.kp ; \
   v5 = v1 + domain.kp ; \
   v6 = v2 + domain.kp ; \
   v7 = v3 + domain.kp ;

#define NDSET2D(v,v1,v2,v3,v4)  \
   v4 = v ;   \
   v1 = v4 + 1 ;  \
   v2 = v1 + domain.jp ;  \
   v3 = v4 + domain.jp ;

#define zabs2(z)    ( real(z)*real(z)+imag(z)*imag(z) )


//
// Domain structure for structured mesh loops
//
struct ADomain
{
   ADomain( LoopLength ilen, Indx_type ndims ) 
      : ndims(ndims), NPNL(2), NPNR(1)
   {
      Indx_type rzmax;
      switch ( ilen ) {
         case LONG : {
            if ( ndims == 2 ) {
               rzmax = 156;
            } else if ( ndims == 3 ) {
               rzmax = 28;
            } 
            break;
         }
         case MEDIUM : {
            if ( ndims == 2 ) {
               rzmax = 64;
            } else if ( ndims == 3 ) {
               rzmax = 16;
            } 
            break;
         }
         case SHORT : {
            if ( ndims == 2 ) {
               rzmax = 8;
            } else if ( ndims == 3 ) {
               rzmax = 4;
            } 
            break;
         }

         default : { }
      }

      imin = NPNL;
      jmin = NPNL;
      imax = rzmax + NPNR;
      jmax = rzmax + NPNR;
      jp = imax - imin + 1 + NPNL + NPNR;

      if ( ndims == 2 ) {
         kmin = 0;
         kmax = 0;
         kp = 0;
         nnalls = jp * (jmax - jmin + 1 + NPNL + NPNR) ;
      } else if ( ndims == 3 ) {
         kmin = NPNL;
         kmax = rzmax + NPNR;
         kp = jp * (jmax - jmin + 1 + NPNL + NPNR);
         nnalls = kp * (kmax - kmin + 1 + NPNL + NPNR) ;
      }

      fpn = 0;
      lpn = nnalls - 1;
      frn = fpn + NPNL * (kp + jp) + NPNL;
      lrn = lpn - NPNR * (kp + jp) - NPNR;

      fpz = frn - jp - kp - 1;
      lpz = lrn;

      real_zones = new Indx_type[nnalls];
      for (Indx_type i = 0; i < nnalls; ++i) real_zones[i] = -1;

      n_real_zones = 0;

      if ( ndims == 2 ) {

         for (Indx_type j = jmin; j < jmax; j++) {
            for (Indx_type i = imin; i < imax; i++) {
               Indx_type ip = i + j*jp ;

               Indx_type id = n_real_zones;
               real_zones[id] = ip;
               n_real_zones++;
            }
         }

      } else if ( ndims == 3 ) {

         for (Indx_type k = kmin; k < kmax; k++) { 
            for (Indx_type j = jmin; j < jmax; j++) {
               for (Indx_type i = imin; i < imax; i++) {
                  Indx_type ip = i + j*jp + kp*k ;

                  Indx_type id = n_real_zones;
                  real_zones[id] = ip;
                  n_real_zones++;
               }
            }
         } 

      }

   }

   ~ADomain() 
   {
      if (real_zones) delete [] real_zones; 
   }

   Indx_type ndims;
   Indx_type NPNL;
   Indx_type NPNR;

   Indx_type imin;
   Indx_type jmin;
   Indx_type kmin;
   Indx_type imax;
   Indx_type jmax;
   Indx_type kmax;

   Indx_type jp;
   Indx_type kp;
   Indx_type nnalls;

   Indx_type fpn;
   Indx_type lpn;
   Indx_type frn;
   Indx_type lrn;

   Indx_type fpz;
   Indx_type lpz;

   Indx_type* real_zones;
   Indx_type  n_real_zones;
};


#endif  // closing endif for header file include guard
